# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import copy
from iva import edge
from pyfastaq import intervals

class TestEdge(unittest.TestCase):
    def test_open_end(self):
        '''test open_end'''
        e1 = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e2 = edge.Edge('c1', 42, 1, 'c2', 1, 10)
        self.assertEqual(e1.open_end('c1'), edge.LEFT)
        self.assertEqual(e1.open_end('c2'), edge.LEFT)
        self.assertEqual(e2.open_end('c1'), edge.RIGHT)
        self.assertEqual(e2.open_end('c2'), edge.RIGHT)


    def test_reverse(self):
        '''test reverse'''
        e = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e.reverse()
        self.assertEqual(e, edge.Edge('c2', 10, 50, 'c1', 42, 1))


    def test_make_contig_forwards(self):
        '''test make_contig_forwards'''
        e = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e._make_contig_forwards('c1')
        self.assertEqual(e, edge.Edge('c1', 1, 42, 'c2', 50, 10))
        e._make_contig_forwards('c2')
        self.assertEqual(e, edge.Edge('c2', 10, 50, 'c1', 42, 1))
        e._make_contig_forwards('c1')
        self.assertEqual(e, edge.Edge('c1', 1, 42, 'c2', 50, 10))


    def test_make_contig_first(self):
        '''test make_contig_first'''
        e = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e_original = copy.copy(e)
        e.make_contig_first('c1')
        self.assertEqual(e, e_original)
        e.make_contig_first('c2')
        self.assertEqual(e, e_original)
        e.make_contig_first('c1')
        self.assertEqual(e, e_original)


    def test_change_hit_coords_with_intersection(self):
        '''test test_change_hit_coords_with_intersection'''
        e = edge.Edge('c1', 1, 42, 'c2', 10, 50)
        e._change_hit_coords_with_intersection('c1', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 20, 30, 'c2', 29, 38))

        e = edge.Edge('c1', 1, 42, 'c2', 10, 50)
        e._change_hit_coords_with_intersection('c2', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 11, 22, 'c2', 20, 30))

        e = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e._change_hit_coords_with_intersection('c1', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 20, 30, 'c2', 31, 22))

        e = edge.Edge('c1', 1, 42, 'c2', 50, 10)
        e._change_hit_coords_with_intersection('c2', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 21, 32, 'c2', 30, 20))

        e = edge.Edge('c1', 42, 1, 'c2', 10, 50)
        e._change_hit_coords_with_intersection('c1', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 30, 20, 'c2', 22, 31))

        e = edge.Edge('c1', 42, 1, 'c2', 10, 50)
        e._change_hit_coords_with_intersection('c2', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 32, 21, 'c2', 20, 30))

        e = edge.Edge('c1', 42, 1, 'c2', 50, 10)
        e._change_hit_coords_with_intersection('c1', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 30, 20, 'c2', 38, 29))

        e = edge.Edge('c1', 42, 1, 'c2', 50, 10)
        e._change_hit_coords_with_intersection('c2', intervals.Interval(20, 30))
        self.assertEqual(e, edge.Edge('c1', 22, 11, 'c2', 30, 20))


    def test_merge_into(self):
        '''test merge_into'''
        e1 = edge.Edge('c1', 71, 100, 'c2', 1, 30)
        e2 = edge.Edge('c2', 41, 80, 'c3', 1, 40)
        e2_original = copy.copy(e2)
        self.assertFalse(e1.merge_into(e2, 'c2'))
        self.assertEqual(e2, e2_original)

        e1 = edge.Edge('c1', 51, 100, 'c2', 1, 50)
        e2 = edge.Edge('c2', 41, 80, 'c3', 1, 40)
        e2_original = copy.copy(e2)
        self.assertTrue(e1.merge_into(e2, 'c2'))
        self.assertEqual(e2, e2_original)
        self.assertEqual(e1, edge.Edge('c1', 91, 100, 'c3', 1, 10))

        e2 = edge.Edge('c1', 51, 100, 'c2', 1, 50)
        e1 = edge.Edge('c2', 41, 80, 'c3', 1, 40)
        e2_original = copy.copy(e2)
        self.assertTrue(e1.merge_into(e2, 'c2'))
        self.assertEqual(e2, e2_original)
        self.assertEqual(e1, edge.Edge('c1', 91, 100, 'c3', 1, 10))

        e1 = edge.Edge('c1', 51, 1, 'c2', 2, 52)
        e2 = edge.Edge('c2', 41, 80, 'c3', 1, 40)
        e2_original = copy.copy(e2)
        self.assertTrue(e1.merge_into(e2, 'c2'))
        self.assertEqual(e2, e2_original)
        self.assertEqual(e1, edge.Edge('c1', 12, 1, 'c3', 1, 12))

        e2 = edge.Edge('c1', 51, 1, 'c2', 2, 52)
        e1 = edge.Edge('c2', 41, 80, 'c3', 1, 40)
        e2_original = copy.copy(e2)
        self.assertTrue(e1.merge_into(e2, 'c2'))
        self.assertEqual(e2, e2_original)
        self.assertEqual(e1, edge.Edge('c1', 12, 1, 'c3', 1, 12))




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
from iva import kmers

class Contig:
    def __init__(self, fasta, verbose=0):
        self.fa = fasta
        self.verbose = verbose
        self.left_kmers = kmers.Kmers(left=True, verbose=verbose)
        self.right_kmers = kmers.Kmers(verbose=verbose)


    def __len__(self):
        return len(self.fa)


    def add_left_kmer(self, kmer):
        self.left_kmers.append(kmer)
        

    def add_right_kmer(self, kmer):
        self.right_kmers.append(kmer)


    def extend(self, min_cov, min_ratio, extend_bases):
        if self.verbose >= 2:
            print('    trying to extend left end ...')
        new_left_seq = self.left_kmers.extension(min_cov, min_ratio, extend_bases)
        if self.verbose >= 2:
            print('    trying to extend right end ...')
        new_right_seq = self.right_kmers.extension(min_cov, min_ratio, extend_bases)
        if self.verbose >= 1:
            print('    new left sequence:', new_left_seq)
            print('    new right sequence:', new_right_seq)
        self.fa.seq = new_left_seq + self.fa.seq + new_right_seq
        self.left_kmers = kmers.Kmers(left=True, verbose=self.verbose)
        self.right_kmers = kmers.Kmers(verbose=self.verbose)
        return len(new_left_seq), len(new_right_seq)



# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import tempfile
import os
import sys
import shutil
import multiprocessing
from iva import mapping, seed
import pyfastaq

class Error (Exception): pass

class SeedProcessor:
    def __init__(self, seeds_fasta, reads1, reads2, outfile, index_k=15, index_s=3, threads=1, max_insert=500, minid=0.9, seed_stop_length=500, extend_length=50, overlap_length=None, ext_min_cov=5, ext_min_ratio=2, verbose=0, seed_length=None, seed_min_count=10, seed_max_count=100000000, kmc_threads=1):
        self.seeds_fasta = seeds_fasta
        self.reads1 = reads1
        self.reads2 = reads2
        self.outfile = outfile
        self.index_k = index_k
        self.index_s = index_s
        self.threads = threads
        self.kmc_threads = kmc_threads
        self.max_insert = max_insert
        self.minid = minid
        self.seed_stop_length = seed_stop_length
        self.extend_length = extend_length
        self.overlap_length = overlap_length
        self.ext_min_cov = ext_min_cov
        self.ext_min_ratio = ext_min_ratio
        self.verbose = verbose
        self.seed_length = seed_length
        self.seed_min_count = seed_min_count
        self.seed_max_count = seed_max_count

        self.original_seeds = {}
        pyfastaq.tasks.file_to_dict(seeds_fasta, self.original_seeds)
        self.processed_seeds = {}
        self.tmpdir = None
        self.bam_file = None


    def _make_new_seed(self, seed_name):
        if self.verbose:
            print('Making new seed for', seed_name, ' ... start')
        tmp_prefix = os.path.join(self.tmpdir, 'out')
        seed_reads = tmp_prefix + '.' + seed_name + '.reads_1.fa'
        if len(self.original_seeds[seed_name]) > self.seed_stop_length:
            start = int(0.5 * len(self.original_seeds[seed_name]) - 0.5 * self.seed_stop_length)
            end = int(0.5 * len(self.original_seeds[seed_name]) + 0.5 * self.seed_stop_length)
        else:
            start = None
            end = None
        if self.verbose:
            print('Making new seed for', seed_name, ' ... getting reads')
        mapping.bam_file_to_region_fasta(self.bam_file, seed_reads, seed_name, start, end)
        if self.verbose:
            print('Making new seed for', seed_name, ' ... finding most common kmer')
        new_seed = seed.Seed(
            extend_length = self.extend_length,
            overlap_length = self.overlap_length,
            reads1 = seed_reads,
            ext_min_cov = self.ext_min_cov,
            ext_min_ratio = self.ext_min_ratio,
            verbose = self.verbose,
            seed_length = self.seed_length,
            seed_min_count = self.seed_min_count,
            seed_max_count = self.seed_max_count,
            kmc_threads = self.kmc_threads,
            map_threads = self.threads
        )
        if len(new_seed) == 0:
            print('Warning: could not get most common kmer for', seed_name)
            return

        if self.verbose:
            print('Making new seed for', seed_name, ' ... extending most common kmer')

        new_seed.extend(self.reads1, self.reads2, self.seed_stop_length)
        f = pyfastaq.utils.open_file_write(tmp_prefix + '.' + seed_name + '.fa')
        print(pyfastaq.sequences.Fasta('seed.' + seed_name, new_seed.seq[10:-10]), file=f)
        pyfastaq.utils.close(f)
        if self.verbose:
            print('Making new seed for', seed_name, ' ... finished')


    def process(self):
        self.tmpdir = tempfile.mkdtemp(prefix='tmp.process_seeds.', dir=os.getcwd())
        tmp_prefix = os.path.join(self.tmpdir, 'out')
        mapping.map_reads(self.reads1, self.reads2, self.seeds_fasta, tmp_prefix,
            index_k = self.index_k,
            index_s = self.index_s,
            threads = self.threads,
            max_insert = self.max_insert,
            minid = self.minid,
            sort = True)
        self.bam_file = tmp_prefix + '.bam'
        threads = min(8, self.threads) # to save peak memory going too high
        threads = self.threads

        if self.verbose:
            print('Processing seeds with', threads, 'threads:', list(self.original_seeds.keys()))

        pool = multiprocessing.Pool(threads)
        pool.map(self._make_new_seed, list(self.original_seeds.keys()))
        pool.close()
        pool.join()
        if self.verbose:
            print('... finished processing seeds')

        new_seeds = {}
        for seed_name in self.original_seeds:
            fname = tmp_prefix + '.' + seed_name + '.fa'
            if os.path.exists(fname):
                pyfastaq.tasks.file_to_dict(fname, new_seeds)

        if len(new_seeds) == 0:
            raise Error('Error! did not make any new seeds. Cannot continue')
        f = pyfastaq.utils.open_file_write(self.outfile)
        for seq in new_seeds.values():
            print(seq, file=f)
        pyfastaq.utils.close(f)
        shutil.rmtree(self.tmpdir)




# Copyright (c) 2014-2016 Genome Research Ltd.
#
# This file is part of IVA.
#
# IVA is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
import pickle
import shutil
import os
import filecmp
import pysam
import pyfastaq
from iva import mapping

modules_dir = os.path.dirname(os.path.abspath(mapping.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


# different smalt version output slightly different BAMs. Some columns
# should never change, so check just those ones
def get_sam_columns(bamfile):
    sams = []
    sam_reader = pysam.Samfile(bamfile, "rb")
    for sam in sam_reader.fetch(until_eof=True):
        if sam.is_unmapped:
            refname = None
        else:
            refname = sam_reader.getrname(sam.tid)
        sams.append((sam.qname, sam.flag, refname, sam.pos, sam.cigar, sam.seq))
    return sams


class TestMapping(unittest.TestCase):
    def test_smalt_in_path(self):
        '''Test that smalt is in the user's path'''
        assert(shutil.which('smalt') is not None)


    def test_smalt_in_path(self):
        '''Test that samtools is in the user's path'''
        assert(shutil.which('samtools') is not None)


    def test_map_reads(self):
        '''Test mapping reads'''
        ref = os.path.join(data_dir, 'mapping_test.ref.trimmed.fa')
        reads_prefix = os.path.join(data_dir, 'mapping_test.reads')
        out_prefix = 'tmp.out'
        mapping.map_reads(reads_prefix + '_1.fastq', reads_prefix + '_2.fastq', ref, out_prefix)
        expected = get_sam_columns(os.path.join(data_dir, 'mapping_test.smalt.out.bam'))
        got = get_sam_columns(out_prefix + '.bam')
        self.assertListEqual(expected, got)
        os.unlink(out_prefix + '.bam')


    def test_map_reads_and_sort(self):
        '''Test mapping reads and sort BAM'''
        ref = os.path.join(data_dir, 'mapping_test.ref.trimmed.fa')
        reads_prefix = os.path.join(data_dir, 'mapping_test.reads')
        out_prefix = 'tmp.out'
        mapping.map_reads(reads_prefix + '_1.fastq', reads_prefix + '_2.fastq', ref, out_prefix, sort=True, verbose=3)
        expected = get_sam_columns(os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam'))
        got = get_sam_columns(out_prefix + '.bam')
        self.assertListEqual(expected, got)
        os.unlink(out_prefix + '.bam')
        os.unlink(out_prefix + '.bam.bai')
        os.unlink(out_prefix + '.unsorted.bam')


    def test_map_reads_wth_flag(self):
        '''Test map_reads with required flag'''
        ref = os.path.join(data_dir, 'mapping_test.ref.trimmed.fa')
        reads_prefix = os.path.join(data_dir, 'mapping_test.reads')
        out_prefix = 'tmp.out'
        mapping.map_reads(reads_prefix + '_1.fastq', reads_prefix + '_2.fastq', ref, out_prefix, required_flag=12, verbose=3)
        expected = get_sam_columns(os.path.join(data_dir, 'mapping_test.smalt.out.flag12.bam'))
        got = get_sam_columns(out_prefix + '.bam')
        self.assertListEqual(expected, got)
        os.unlink(out_prefix + '.bam')


    def test_get_bam_region_coverage_rev(self):
        '''Test get_bam_region_coverage reverse strand'''
        bam = os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam')
        cov = mapping.get_bam_region_coverage(bam, 'ref', 190, rev=True, verbose=3)
        f = open(os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam.rev.cov'), 'rb')
        expected = pickle.load(f)
        f.close()
        self.assertListEqual(cov, expected)


    def test_get_bam_region_coverage_fwd(self):
        '''Test get_bam_region_coverage forward strand'''
        bam = os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam')
        cov = mapping.get_bam_region_coverage(bam, 'ref', 190, verbose=3)
        f = open(os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam.fwd.cov'), 'rb')
        expected = pickle.load(f)
        f.close()
        self.assertListEqual(cov, expected)


    def test_get_bam_region_coverage_fwd_And_rev(self):
        '''Test get_bam_region_coverage both strands'''
        bam = os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam')
        cov = mapping.get_bam_region_coverage(bam, 'ref', 190, verbose=3, both_strands=True)
        f = open(os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam.fwd_and_rev.cov'), 'rb')
        expected = pickle.load(f)
        f.close()
        self.assertListEqual(cov, expected)
       

    def test_remove_indels(self):
        '''Test _remove_indels'''
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('ac+1Xgt'), '+')))
        self.assertEqual('ac+1Xgt', ''.join(mapping._remove_indels(list('ac+1Xgt'), '-')))
        self.assertEqual('ac-1Xgt', ''.join(mapping._remove_indels(list('ac-1Xgt'), '+')))
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('ac+2XXgt'), '+')))
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('ac+10XXXXXXXXXXgt'), '+')))
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('ac-10XXXXXXXXXXgt'), '-')))
        self.assertEqual('a-1Xcgt', ''.join(mapping._remove_indels(list('a-1Xc+1Xg+10XXXXXXXXXXt'), '+')))
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('+1Xacgt'), '+')))
        self.assertEqual('acgt', ''.join(mapping._remove_indels(list('acgt+1X'), '+')))


    def test_strip_mpileup_coverage_string(self):
        '''Test strip_mpileup_coverage_string'''
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('acg^[t'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('acgt$'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('*ac*gt'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('*a$c^[gt$'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('ac+1Xgt'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('acg+10XXXXXXXXXXt'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('ac-1Xgt'))
        self.assertEqual('acgt', mapping.strip_mpileup_coverage_string('acg-10XXXXXXXXXXt'))
        self.assertEqual('aaa', mapping.strip_mpileup_coverage_string('a-1Na^+a'))


    def test_consensus_base(self):
        '''Test consensus_base'''
        keys = ['A', 'C', 'G', 'T']
        self.assertEqual(None, mapping.consensus_base({}, keys))
        self.assertEqual('G', mapping.consensus_base({'A': 2, 'C': 2, 'G': 4}, keys, ratio=0.5))
        self.assertEqual('G', mapping.consensus_base({'A': 2, 'C': 1, 'G': 4, 'T':1}, keys, ratio=0.5))
        self.assertEqual(None, mapping.consensus_base({'A': 2, 'C': 1, 'G': 4, 'T':2}, keys, ratio=0.5))
        self.assertEqual('G', mapping.consensus_base({'A': 2, 'C': 1, 'G': 4, 'T':2}, keys, ratio=0.43))


    def test_consensus_base_both_strands(self):
        '''Test consensus_base_both_strands'''
        forward_keys = set(['A', 'C', 'G', 'T', 'N'])
        reverse_keys = set(['a', 'c', 'g', 't', 'n'])
        counts = [
            ({}, None),
            ({'A': 2, 'C': 2, 'G': 4, 'a': 2, 'c': 2, 'g': 5}, 'G'),
            ({'A': 2, 'C': 2, 'G': 3, 'a': 2, 'c': 2, 'g': 5}, None),
            ({'A': 2, 'C': 2, 'G': 4, 'a': 2, 'c': 2, 'g': 3}, None),
        ]

        for counts_dict, expected in counts:
            self.assertEqual(expected, mapping.consensus_base_both_strands(counts_dict, forward_keys, reverse_keys, ratio=0.5))


    def test_find_incorrect_ref_bases(self):
        '''Test find_incorrect_ref_bases'''
        bam = os.path.join(data_dir, 'mapping_test.find_incorrect_ref_bases.bam')
        ref = os.path.join(data_dir, 'mapping_test.find_incorrect_ref_bases.fasta')
        bad_bases = mapping.find_incorrect_ref_bases(bam, ref)
        expected = {'1': [(197, 'A', 'T'), (280, 'T', 'G')]}
        self.assertTrue(expected, bad_bases)


    def test_soft_clipped(self):
        '''Test soft_clipped'''
        expected = [
            (5, 0),
            (0, 0),
            (0, 0),
            (0, 5),
            (0, 0),
            None,
            (0, 0),
            (0, 0),
            (2, 0),
            (0, 1),
            None,
            None,
            (1, 1),
            (0, 1)
        ]

        sam_reader = pysam.Samfile(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), "rb")
        i = 0
        for sam in sam_reader.fetch(until_eof=True):
            self.assertEqual(mapping.soft_clipped(sam), expected[i])
            i += 1


    def test_sam_to_fasta(self):
        '''Test sam_to_fasta'''
        expected_seqs = {}
        pyfastaq.tasks.file_to_dict(os.path.join(data_dir, 'mapping_test.reads_1.fasta'), expected_seqs)
        pyfastaq.tasks.file_to_dict(os.path.join(data_dir, 'mapping_test.reads_2.fasta'), expected_seqs)
        sam_reader = pysam.Samfile(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), "rb")
        for sam in sam_reader.fetch(until_eof=True):
            fa = mapping.sam_to_fasta(sam)
            self.assertTrue(fa.id in expected_seqs)
            self.assertEqual(fa, expected_seqs[fa.id])


    def test_can_extend(self):
        '''Test can_extend'''
        expected = [
            (True, False),
            (False, False),
            (False, False),
            (False, True),
            (False, False),
            (False, False),
            (False, False),
            (False, False),
            (True, False),
            (False, False),
            (False, False),
            (False, False),
            (False, False),
            (False, False),

        ]

        sam_reader = pysam.Samfile(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), "rb")
        i = 0
        for sam in sam_reader.fetch(until_eof=True):
            self.assertEqual(mapping._can_extend(sam, 190, min_clip=2), expected[i])
            i += 1


    def test_get_pair_type(self):
        '''Test get_pair_type'''
        expected = [
            (mapping.CAN_EXTEND_LEFT, mapping.KEEP),
            (mapping.KEEP, mapping.CAN_EXTEND_RIGHT),
            (mapping.KEEP, mapping.KEEP),
            (mapping.NOT_USEFUL, mapping.NOT_USEFUL),
            (mapping.CAN_EXTEND_LEFT, mapping.KEEP),
            (mapping.BOTH_UNMAPPED, mapping.BOTH_UNMAPPED),
            (mapping.NOT_USEFUL, mapping.NOT_USEFUL)
        ]

        sam_reader = pysam.Samfile(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), "rb")
        previous_sam = None
        i = 0
        for sam in sam_reader.fetch(until_eof=True):
            if previous_sam is None:
                previous_sam = sam
                continue

            types = mapping.get_pair_type(previous_sam, sam, 190, 1000, min_clip=2)
            self.assertEqual(types, expected[i])
            i += 1
            previous_sam = None


    def test_get_ref_name(self):
        '''Test get_ref_name'''
        expected = ['ref'] * 14
        for i in ([5,10,11]):
            expected[i] = None
        sam_reader = pysam.Samfile(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), "rb")
        i = 0
        for sam in sam_reader.fetch(until_eof=True):
            self.assertEqual(mapping.get_ref_name(sam, sam_reader), expected[i])
            i += 1


    def test_bam_file_to_fasta_pair_files(self):
        '''Test bam_file_to_fasta_pair_files'''
        tmp1 = 'tmp.to_fasta_1.fa'
        tmp2 = 'tmp.to_fasta_2.fa'
        mapping.bam_file_to_fasta_pair_files(os.path.join(data_dir, 'mapping_test.smalt.out.bam'), tmp1, tmp2)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.reads_1.fasta'), tmp1))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.reads_2.fasta'), tmp2))
        os.unlink(tmp1)
        os.unlink(tmp2)


    def test_bam_file_to_fasta_pair_files_region(self):
        '''Test bam_file_to_fasta_pair_files with a region'''
        tmp1 = 'tmp.to_fasta_1.fa'
        tmp2 = 'tmp.to_fasta_2.fa'
        mapping.bam_file_to_fasta_pair_files(os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam'), tmp1, tmp2, chromosome='ref', start=25, end=150)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.bam_to_region_1.fa'), tmp1))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.bam_to_region_2.fa'), tmp2))
        os.unlink(tmp1)
        os.unlink(tmp2)


    def test_bam_file_to_region_fasta(self):
        '''Test bam_file_to_region_fasta'''
        tmp = 'tmp.to_fasta.fa'
        bam = os.path.join(data_dir, 'mapping_test.smalt.out.sorted.bam')
        mapping.bam_file_to_region_fasta(bam, tmp, 'ref', start=42, end=142)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'mapping_test.bam_to_region.fasta'), tmp))
        os.unlink(tmp)


    def test_bam_to_fasta(self):
        '''Test bam_to_fasta'''
        # TODO
        pass


    def test_total_ref_length_from_bam(self):
        '''Test _total_ref_length_from_bam'''
        bam = os.path.join(data_dir, 'mapping_test.total_ref_length_from_bam.bam')
        self.assertEqual(300, mapping._total_ref_length_from_bam(bam))


    def _mean_read_length(self):
        '''Test _mean_read_length'''
        bam = os.path.join(data_dir, 'mapping_test.mean_read_length.bam')
        lengths = [19, 18, 20, 17, 20, 20, 20, 20]
        self.assertEqual(19, mapping._mean_read_length(bam, head=1))
        self.assertEqual(18, mapping._mean_read_length(bam, head=2))
        self.assertEqual(int(sum(lengths) / len(lengths)), mapping._mean_read_length(bam))

